"""Promote a P router to a PE router."""

import json
from typing import Any

from orchestrator.config.assignee import Assignee
from orchestrator.forms import FormPage
from orchestrator.forms.validators import Label
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, UUIDstr
from orchestrator.utils.errors import ProcessFailureError
from orchestrator.utils.json import json_dumps
from orchestrator.workflow import StepList, begin, done, inputstep, step, workflow
from orchestrator.workflows.steps import resync, store_process_subscription, unsync
from orchestrator.workflows.utils import wrap_modify_initial_input_form
from pydantic import ConfigDict, model_validator

from gso.products.product_blocks.router import RouterRole
from gso.products.product_types.router import Router
from gso.services import lso_client
from gso.services.kentik_client import KentikClient, NewKentikDevice
from gso.services.lso_client import lso_interaction
from gso.services.subscriptions import get_all_active_sites
from gso.utils.helpers import generate_inventory_for_active_routers
from gso.utils.shared_enums import Vendor
from gso.utils.types import TTNumber


def initial_input_form_generator(subscription_id: UUIDstr) -> FormGenerator:
    """Promote P router to PE router."""
    subscription = Router.from_subscription(subscription_id)

    class PromotePToPEForm(FormPage):
        model_config = ConfigDict(title=f"Promote {subscription.router.router_fqdn} to PE router?")

        tt_number: TTNumber

        @model_validator(mode="before")
        def router_must_be_nokia_p(cls, data: Any) -> Any:
            if not (subscription.router.router_role == RouterRole.P and subscription.router.vendor == Vendor.NOKIA):
                msg = "PE promotion workflow can only be run on Nokia P routers."
                raise ValueError(msg)

            return data

    user_input = yield PromotePToPEForm

    return user_input.model_dump() | {"subscription": subscription}


@step("Evacuate the router by setting isis_overload")
def set_isis_overload(subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr) -> None:
    """Evacuate the router by setting isis overload."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - Set ISIS overload",
        "verb": "set_isis_overload",
    }

    lso_client.execute_playbook(
        playbook_name="promote_p_to_pe.yaml",
        callback_route=callback_route,
        inventory=subscription["router"]["router_fqdn"],
        extra_vars=extra_vars,
    )


@step("[DRY RUN] Deploy PE base config")
def deploy_pe_base_config_dry(
    subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr
) -> None:
    """Perform a dry run of adding the base config to the router."""
    extra_vars = {
        "dry_run": True,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - deploy PE base config",
        "verb": "deploy_pe_base_config",
        "pe_router_list": generate_inventory_for_active_routers(RouterRole.PE)["all"]["hosts"],
        "geant_sites": json.loads(json_dumps(get_all_active_sites())),
    }

    lso_client.execute_playbook(
        playbook_name="promote_p_to_pe.yaml",
        callback_route=callback_route,
        inventory=subscription["router"]["router_fqdn"],
        extra_vars=extra_vars,
    )


@step("[FOR REAL] Deploy PE base config")
def deploy_pe_base_config_real(
    subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr
) -> None:
    """Perform a real run of adding the base config to the router."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - deploy PE base config",
        "verb": "deploy_pe_base_config",
        "pe_router_list": generate_inventory_for_active_routers(RouterRole.PE)["all"]["hosts"],
        "geant_sites": json.loads(json_dumps(get_all_active_sites())),
    }

    lso_client.execute_playbook(
        playbook_name="promote_p_to_pe.yaml",
        callback_route=callback_route,
        inventory=subscription["router"]["router_fqdn"],
        extra_vars=extra_vars,
    )


@inputstep("Prompt EARL insertion", assignee=Assignee.SYSTEM)
def prompt_insert_in_earl(subscription: dict[str, Any]) -> FormGenerator:
    """Wait for confirmation from an operator that the router has been inserted in EARL."""

    class EARLPrompt(FormPage):
        model_config = ConfigDict(title="Update RADIUS clients")

        info_label: Label = f"Please add the router {subscription["router"]["router_fqdn"]} to EARL."

    yield EARLPrompt

    return {}


@step("Create Kentik device")
def create_kentik_device(subscription: Router) -> State:
    """Create a new device in Kentik."""
    if not (
        subscription.router.router_site
        and subscription.router.router_site.site_name
        and subscription.router.router_site.site_tier
        and subscription.router.router_fqdn
    ):
        msg = "Router object is missing required properties."
        raise ProcessFailureError(msg)

    kentik_client = KentikClient()
    kentik_site = kentik_client.get_site_by_name(subscription.router.router_site.site_name)

    if not kentik_site:
        msg = "Site could not be found in Kentik."
        raise ProcessFailureError(msg, details=subscription.router.router_site.site_name)

    new_device = NewKentikDevice(
        device_name=subscription.router.router_fqdn,
        device_description=str(subscription.subscription_id),
        sending_ips=[str(subscription.router.router_lo_ipv4_address)],
        site_id=kentik_site["id"],
        device_snmp_ip=str(subscription.router.router_lo_ipv4_address),
        device_bgp_flowspec=False,
        device_bgp_neighbor_ip=str(subscription.router.router_lo_ipv4_address),
        device_bgp_neighbor_ip6=str(subscription.router.router_lo_ipv6_address),
    )
    kentik_device = kentik_client.create_device(new_device)

    return {"kentik_device": kentik_device}


@step("[DRY RUN] Include new PE into SDP mesh on other Nokia PEs")
def update_sdp_mesh_dry(subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr) -> None:
    """Perform a dry run for updating the SDP mesh with the new router."""
    extra_vars = {
        "dry_run": True,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - "
        f"Update the SDP mesh for L2circuits(epipes) config on PE NOKIA routers",
        "verb": "update_sdp_mesh",
        "pe_router_list": {
            subscription["router"]["router_fqdn"]: {
                "lo4": str(subscription["router"]["router_lo_ipv4_address"]),
                "lo6": str(subscription["router"]["router_lo_ipv6_address"]),
            }
        },
    }

    lso_client.execute_playbook(
        playbook_name="update_pe_sdp_mesh.yaml",
        callback_route=callback_route,
        inventory=generate_inventory_for_active_routers(router_role=RouterRole.PE, router_vendor=Vendor.NOKIA),
        extra_vars=extra_vars,
    )


@step("[FOR REAL] Include new PE into SDP mesh on other Nokia PEs")
def update_sdp_mesh_real(
    subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr
) -> None:
    """Update the SDP mesh for L2 circuits(epipes) config on PE NOKIA routers."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - "
        f"Update the SDP mesh for l2circuits(epipes) config on PE NOKIA routers",
        "verb": "update_sdp_mesh",
        "pe_router_list": {
            subscription["router"]["router_fqdn"]: {
                "lo4": str(subscription["router"]["router_lo_ipv4_address"]),
                "lo6": str(subscription["router"]["router_lo_ipv6_address"]),
            }
        },
    }

    lso_client.execute_playbook(
        playbook_name="update_pe_sdp_mesh.yaml",
        callback_route=callback_route,
        inventory=generate_inventory_for_active_routers(router_role=RouterRole.PE, router_vendor=Vendor.NOKIA),
        extra_vars=extra_vars,
    )


@step("[DRY RUN] Remove P from all PEs")
def remove_p_from_pe_dry(
    subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr
) -> None:
    """Perform a dry run of removing the P router from all the PE routers."""
    extra_vars = {
        "dry_run": True,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - "
        f"Remove P-only router neighbour {subscription["router"]["router_fqdn"]} from PE routers P-only group",
        "verb": "remove_p_from_pe",
    }

    lso_client.execute_playbook(
        playbook_name="update_ibgp_mesh.yaml",
        callback_route=callback_route,
        inventory=generate_inventory_for_active_routers(RouterRole.PE),
        extra_vars=extra_vars,
    )


@step("[FOR REAL] Remove P from all PEs")
def remove_p_from_pe_real(
    subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr
) -> None:
    """Remove the P router from all the PE routers."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - "
        f"Remove P-only router neighbour {subscription["router"]["router_fqdn"]} from PE routers P-only group",
        "verb": "remove_p_from_pe",
    }

    lso_client.execute_playbook(
        playbook_name="update_ibgp_mesh.yaml",
        callback_route=callback_route,
        inventory=generate_inventory_for_active_routers(RouterRole.PE),
        extra_vars=extra_vars,
    )


@step("[DRY RUN] Add PE mesh to PE")
def add_pe_mesh_to_pe_dry(
    subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr
) -> None:
    """Perform a dry run of adding list of PE routers into iGEANT/iGEANT6 of promoted router."""
    extra_vars = {
        "dry_run": True,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - "
        f"Add list of PE routers into iGEANT/iGEANT6 of promoted router",
        "verb": "add_pe_mesh_to_pe",
        "pe_router_list": generate_inventory_for_active_routers(RouterRole.PE)["all"]["hosts"],
    }

    lso_client.execute_playbook(
        playbook_name="update_ibgp_mesh.yaml",
        callback_route=callback_route,
        inventory=subscription["router"]["router_fqdn"],
        extra_vars=extra_vars,
    )


@step("[FOR REAL] Add PE mesh to PE")
def add_pe_mesh_to_pe_real(
    subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr
) -> None:
    """Perform a real run of adding list of PE routers into iGEANT/iGEANT6 of promoted router."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - "
        f"Add list of PE routers into iGEANT/iGEANT6 of promoted router",
        "verb": "add_pe_mesh_to_pe",
        "pe_router_list": generate_inventory_for_active_routers(RouterRole.PE)["all"]["hosts"],
    }

    lso_client.execute_playbook(
        playbook_name="update_ibgp_mesh.yaml",
        callback_route=callback_route,
        inventory=subscription["router"]["router_fqdn"],
        extra_vars=extra_vars,
    )


@step("[DRY RUN] Add PE to PE mesh")
def add_pe_to_pe_mesh_dry(
    subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr
) -> None:
    """Perform a dry run of adding the promoted router to all PE routers in iGEANT/iGEANT6."""
    extra_vars = {
        "dry_run": True,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - "
        f"Add promoted router to all PE routers in iGEANT/iGEANT6.",
        "verb": "add_pe_to_pe_mesh",
    }

    lso_client.execute_playbook(
        playbook_name="update_ibgp_mesh.yaml",
        callback_route=callback_route,
        inventory=generate_inventory_for_active_routers(RouterRole.PE),
        extra_vars=extra_vars,
    )


@step("[FOR REAL] Add PE to PE mesh")
def add_pe_to_pe_mesh_real(
    subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr
) -> None:
    """Perform a real run of adding the promoted router to all PE routers in iGEANT/iGEANT6."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - "
        f"Add promoted router to all PE routers in iGEANT/iGEANT6.",
        "verb": "add_pe_to_pe_mesh",
    }

    lso_client.execute_playbook(
        playbook_name="update_ibgp_mesh.yaml",
        callback_route=callback_route,
        inventory=generate_inventory_for_active_routers(RouterRole.PE),
        extra_vars=extra_vars,
    )


@step("Check iBGP session")
def check_pe_ibgp(subscription: dict[str, Any], callback_route: str) -> None:
    """Check the iBGP session."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "verb": "check_pe_ibgp",
    }

    lso_client.execute_playbook(
        playbook_name="check_ibgp.yaml",
        callback_route=callback_route,
        inventory=subscription["router"]["router_fqdn"],
        extra_vars=extra_vars,
    )


@step("[DRY RUN] Deploy routing instances")
def deploy_routing_instances_dry(
    subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr
) -> None:
    """Perform a dry run of deploying routing instances."""
    extra_vars = {
        "dry_run": True,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - Deploy PE base config",
        "verb": "deploy_routing_instances",
    }

    lso_client.execute_playbook(
        playbook_name="promote_p_to_pe.yaml",
        callback_route=callback_route,
        inventory=subscription["router"]["router_fqdn"],
        extra_vars=extra_vars,
    )


@step("[FOR REAL] Deploy routing instances")
def deploy_routing_instances_real(
    subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr
) -> None:
    """Perform a real run of deploying routing instances."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - Deploy PE base config",
        "verb": "deploy_routing_instances",
    }

    lso_client.execute_playbook(
        playbook_name="promote_p_to_pe.yaml",
        callback_route=callback_route,
        inventory=subscription["router"]["router_fqdn"],
        extra_vars=extra_vars,
    )


@step("Check L3 services")
def check_l3_services(subscription: dict[str, Any], callback_route: str) -> None:
    """Check L3 services."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "verb": "check_base_ris",
    }

    lso_client.execute_playbook(
        playbook_name="check_l3_services.yaml",
        callback_route=callback_route,
        inventory=subscription["router"]["router_fqdn"],
        extra_vars=extra_vars,
    )


@step("Remove ISIS overload")
def remove_isis_overload(
    subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr
) -> None:
    """Remove ISIS overload."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - Remove ISIS overload",
        "verb": "remove_isis_overload",
    }

    lso_client.execute_playbook(
        playbook_name="promote_p_to_pe.yaml",
        callback_route=callback_route,
        inventory=subscription["router"]["router_fqdn"],
        extra_vars=extra_vars,
    )


@step("Set router role to PE (Update subscription model)")
def update_subscription_model(subscription: Router) -> State:
    """Update the subscription model to set router role to PE."""
    subscription.router.router_role = RouterRole.PE

    return {"subscription": subscription}


@step("[DRY RUN] Add all P to PE")
def add_all_p_to_pe_dry(subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr) -> None:
    """Perform a dry run of adding all P routers to the PE router."""
    extra_vars = {
        "dry_run": True,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - Add all P-routers to this new PE",
        "verb": "add_all_p_to_pe",
        "p_router_list": generate_inventory_for_active_routers(RouterRole.P)["all"]["hosts"],
    }

    lso_client.execute_playbook(
        playbook_name="update_ibgp_mesh.yaml",
        callback_route=callback_route,
        inventory=subscription["router"]["router_fqdn"],
        extra_vars=extra_vars,
    )


@step("[FOR REAL] Add all P to PE")
def add_all_p_to_pe_real(
    subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr
) -> None:
    """Perform a real run of adding all P routers to the PE router."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - Add all P-routers to this new PE",
        "verb": "add_all_p_to_pe",
        "p_router_list": generate_inventory_for_active_routers(RouterRole.P)["all"]["hosts"],
    }

    lso_client.execute_playbook(
        playbook_name="update_ibgp_mesh.yaml",
        callback_route=callback_route,
        inventory=subscription["router"]["router_fqdn"],
        extra_vars=extra_vars,
    )


@step("[DRY RUN] Add PE to all P")
def add_pe_to_all_p_dry(subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr) -> None:
    """Perform a dry run of adding promoted router to all PE routers in iGEANT/iGEANT6."""
    extra_vars = {
        "dry_run": True,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - "
        f"Add promoted router to all PE routers in iGEANT/iGEANT6",
        "verb": "add_pe_to_all_p",
    }

    lso_client.execute_playbook(
        playbook_name="update_ibgp_mesh.yaml",
        callback_route=callback_route,
        inventory=generate_inventory_for_active_routers(RouterRole.P),
        extra_vars=extra_vars,
    )


@step("[FOR REAL] Add PE to all P")
def add_pe_to_all_p_real(
    subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr
) -> None:
    """Perform a real run of adding promoted router to all PE routers in iGEANT/iGEANT6."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - "
        f"Add promoted router to all PE routers in iGEANT/iGEANT6",
        "verb": "add_pe_to_all_p",
    }

    lso_client.execute_playbook(
        playbook_name="update_ibgp_mesh.yaml",
        callback_route=callback_route,
        inventory=generate_inventory_for_active_routers(RouterRole.P),
        extra_vars=extra_vars,
    )


@step("[DRY RUN] Delete default routes")
def delete_default_routes_dry(
    subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr
) -> None:
    """Perform a dry run of deleting the default routes."""
    extra_vars = {
        "dry_run": True,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - "
        f"Delete static default routes (part of P base-config)",
        "verb": "delete_default_routes",
    }

    lso_client.execute_playbook(
        playbook_name="promote_p_to_pe.yaml",
        callback_route=callback_route,
        inventory=subscription["router"]["router_fqdn"],
        extra_vars=extra_vars,
    )


@step("[FOR REAL] Delete default routes")
def delete_default_routes_real(
    subscription: dict[str, Any], callback_route: str, tt_number: str, process_id: UUIDstr
) -> None:
    """Perform a real run of deleting the default routes."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - "
        f"Delete static default routes (part of P base-config)",
        "verb": "delete_default_routes",
    }

    lso_client.execute_playbook(
        playbook_name="promote_p_to_pe.yaml",
        callback_route=callback_route,
        inventory=subscription["router"]["router_fqdn"],
        extra_vars=extra_vars,
    )


@workflow(
    "Promote P router to PE router",
    initial_input_form=wrap_modify_initial_input_form(initial_input_form_generator),
    target=Target.MODIFY,
)
def promote_p_to_pe() -> StepList:
    """Promote a P router to a PE router."""
    return (
        begin
        >> store_process_subscription(Target.MODIFY)
        >> unsync
        >> lso_interaction(set_isis_overload)
        >> lso_interaction(deploy_pe_base_config_dry)
        >> lso_interaction(deploy_pe_base_config_real)
        >> prompt_insert_in_earl
        >> create_kentik_device
        >> lso_interaction(remove_p_from_pe_dry)
        >> lso_interaction(remove_p_from_pe_real)
        >> lso_interaction(add_pe_mesh_to_pe_dry)
        >> lso_interaction(add_pe_mesh_to_pe_real)
        >> lso_interaction(add_pe_to_pe_mesh_dry)
        >> lso_interaction(add_pe_to_pe_mesh_real)
        >> lso_interaction(check_pe_ibgp)
        >> lso_interaction(deploy_routing_instances_dry)
        >> lso_interaction(deploy_routing_instances_real)
        >> lso_interaction(check_l3_services)
        >> lso_interaction(update_sdp_mesh_dry)
        >> lso_interaction(update_sdp_mesh_real)
        >> lso_interaction(add_all_p_to_pe_dry)
        >> lso_interaction(add_all_p_to_pe_real)
        >> lso_interaction(add_pe_to_all_p_dry)
        >> lso_interaction(add_pe_to_all_p_real)
        >> lso_interaction(delete_default_routes_dry)
        >> lso_interaction(delete_default_routes_real)
        >> lso_interaction(remove_isis_overload)
        >> update_subscription_model
        >> resync
        >> done
    )
